clear all
clc
rng(1)
%% Parameters 
    
load('RUSK_zc_sounder.mat')

K                   = 8;                                        %number DOAs
N_TX                = 8;                                        %num transmitters
N_RX                = 12;                                       %num receivers
N                   = 2^10;                                     %number time samples in sounding sequence
L                   = 5;                                        %snapshots per DOA estimation
DOA_subcarrier      = 700;                                      %subcarrier to perform DOA estimation
MAX_INDEX           = 10e6;                                      %stop index on rx data (to avoid processing GBs)
STRT_IDX            = 1;                                        %first data sample to observe
CFO_CORRECTION      = 1;                                        %toggle to correct for CFO
SAMPLE_CORRECTION   = 1;                                        %toggle to correct for sample start index
PHASE_CORRECTION    = 1;                                        %toggle to correct random phase offsets for each rx channel pair
PLOT_DET            = 0;                                        %plot toggle to see detections (use for debugging)
PLOT_STS            = 0;                                        %plot toggle to see STS detection (use for debugging)
PHASE_PLOT          = 0;                                        %plot toggle to see phase of each rx clock
CHANNEL_PLOT        = 0;                                        %plot toggle to observe the channel (use for debugging)

%% Load received data
for j = 1:N_RX
    filename        = strcat('<path_to_data_folder>/RUSK_rx/RUSK_rx_ch',num2str(j),'.bin'); % edit this line!            %get address of binary rx data files
    fileID          = fopen(filename);                                                                      %open files
    Q               = fread(fileID,'float32','ieee-le');                                                    %read files as IEEE little endian, float 32 format
    a               = Q(1:2:length(Q));                                                                     %inphase channel
    b               = Q(2:2:length(Q));                                                                     %quadrature channel
    IQ_temp         = [complex(double(a),double(b))];                                                       %convert to complex numbers
    IQ_temp         = IQ_temp(STRT_IDX:end);                                                                %cut off beggining values
    if j == 1 
        IQ          = zeros(N_RX,min([length(IQ_temp),MAX_INDEX]));                                         %if this is the first 
    end
    if length(IQ_temp) >= size(IQ,2)                %if it's too long shave some off
        IQ_temp     = IQ_temp(1:size(IQ,2));
    end
    if length(IQ_temp) < size(IQ,2)                 %if it's too short add some zeros
        IQ_temp     = [IQ_temp;zeros(size(IQ,2)-length(IQ_temp),1)];
    end
    IQ(j,:)         = transpose(IQ_temp);
    fclose(fileID);
end
clear IQ_temp a b filename fileID Q

%% Detect starting frames
max_cap             = round(size(IQ,2)/interval);                       %largest number of captures possible
inds                = zeros(N_RX,max_cap);                              %allocate space for starting indices of detected signals
tx_ids              = zeros(N_RX,max_cap);                              %allocate space for transmitter ID for each detected signal
cfos                = zeros(N_RX,max_cap);                              %store CFO correction factor for each detection

intervals           = interval.*[0:3];
tmp                 = zeros(1,interval);
tmp(1:length(zc))   = zc;
zc_full             = repmat(tmp,1,4);

for k = 1:N_RX
    
    [detector,lags]     = xcorr(IQ(k,:),zc);                            %detect signals by correlating rx data with ZC sequence
    detector            = abs(detector./max(abs(detector)));            %normalize
    d_thresh            = 75*mean(detector);                            %arbitrary noise threshold
    detector(detector<d_thresh) = 0;                                    %shave off noise
    detector(lags<=0)   = 0;                                            %fix errors from autocorr zero padding
    detector            = detector(size(IQ,2)-1:end-interval);          %remove zero padded part of autocorr vector
    [~,ind]             = findpeaks(detector);                          %find detections
    ind                 = sort(ind,'ascend');                           %order detections by time of arrival
    %% PLOT
    if PLOT_DET
        [detector2,lags]     = xcorr(IQ(k,:),zc);
        detector2            = abs(detector2./max(abs(detector2)));
        detector2(lags<=0)   = 0;
        detector2            = detector2(size(IQ,2)-1:end-interval);     
        figure(1)
        clf
        subplot(2,1,1)
        plot(abs(detector))
        grid on
        subplot(2,1,2)
        plot(abs(detector2))
        grid on
    end
    
    %% Sort out multipath detections
    tmp                 = zeros(1,length(ind));                         %temporary array
    region              = 10;                                           %number of samples on each side of detection to observe
    for i = 1:length(ind)                                               %for each packet detection
        if ind(i)>0                                                     %if we haven't already nulled this detection
            val             = ind(i);                                   %store detection sample index
            tmpind          = abs(ind-val);                             %get relative displacement of all other detections
            [~,candidates]  = find(tmpind<region);                      %keep indicies of detections in the nearby region
            [~,maxind]      = max(detector(ind(candidates)));           %find index of largest correlation, this is considered the true arrival 
            candidates(maxind) = [];                                    %remove true arrival from candidates vector
            ind(candidates) = 0;                                        %set all remaining detections to zero
        end
    end
    ind(ind==0)             = [];                                       %remove zeroed out detections
    inds(k,1:length(ind))   = ind;                                      %store in matrix to contain all detections over all rx channels                        
    
    %% Determine which Tx is talking for each packet detection
    for i = 1:length(ind)
        aoi                 = IQ(k,ind(i)+length(zc):ind(i) + length(zc)+45*length(sts_t)+1);   %area of interest
        det                 = sliding_corr(aoi,32);                     %autocorrelate
        det                 = abs(det./max(abs(det)));                  %normalize
        det(det<.7)         = 0;                                        %remove values below detection threshold
        [~,tmp]             = findpeaks(det);                           %find peaks
        num_det             = round(nnz(det)/length(sts_t));            %number of detections
        
        switch num_det                          %logic based on number of STS detected. 
            case {0,1,2,3,4}                    %Tx 1 sent 5 STS
                tx_ids(k,i) = 1;
            case {5,6,7,8,9}                    %Tx 2 sent 10 STS
                tx_ids(k,i) = 2; 
            case {10,11,12,13,14}               %Tx 3 sent 15 STS
                tx_ids(k,i) = 3;
            case {15,16,17,18,19}               %Tx 4 sent 20 STS
                tx_ids(k,i) = 4;
            case {20,21,22,23,24}               %Tx 5 sent 25 STS
                tx_ids(k,i) = 5;
            case {2526,27,28,29}                %Tx 6 sent 30 STS
                tx_ids(k,i) = 6;
            case {30,31,32,33,34}               %Tx 7 sent 35 STS
                tx_ids(k,i) = 7;
            case {35,36,37,38,39,40}            %Tx 8 sent 40 STS
                tx_ids(k,i) = 8;
        end
        
    %% CFO estimation
        if CFO_CORRECTION
            [CFO,cfo_ests]  = est_CFO(aoi,length(sts_t),num_det+1);         %calculate CFO
            cfos(k,i)       = CFO;                                          %record CFO
        end
        %% PLOT
        if PLOT_STS
            figure(2)
            subplot(2,1,1)
            plot(abs(det))
            grid on
            axis([-10,45*length(sts_t),0,1.15])
            title('Autocorrelation at detection start index')
            xlabel('Lag (samples)')
            ylabel('Correlation magnitude')
            subplot(2,1,2)
            zplane(cfo_ests./mean(abs(cfo_ests)))
            grid on
            title('CFO estimates')
        end
        q=1;
    end
end

%% Go over detections, correct errors, insert detections below noise floor

for j = 1:size(tx_ids,2)                                                                                                %for all detections
    true_arrival            = median(tx_ids(:,j));                                                                      %democratic vote on which arrival is correct
    if true_arrival == 0                                                                                                %if no arrival
        last_idx            = j-1;                                                                                      %this is the last informational index value in tx_ids, inds, and cfos
        break                                                                                                           %stop
    end
    for i = 1:N_RX                                                                                                      %for each rx channel
        est_arrival         = tx_ids(i,j);                                                                              %estimated tx channel 
        if est_arrival > true_arrival                                                                                   %when estimated tx channel > true tx channel
            shift_right     = est_arrival-true_arrival;                                                                 %find how many spaces we'll need to shift right
            if j == 1                                                                                                   %special logic for very first detection
                tx_ids(i,:) = [(1:shift_right)+(true_arrival-1),tx_ids(i,1:end-shift_right)];                           %shift tx_ids right by 'shift_right' number of indicies
                inds(i,:)   = [interval*(-1:-1*shift_right)+inds(i,1),inds(i,1:end-shift_right)];                       %shift inds right by 'shift right' number of indicies
                for k = 1:shift_right                                                                                   %deal with CFO estimation
%                     aoi         = IQ(i, inds(i,k)+length(zc)+1 : inds(i,k)+length(zc)+45*length(sts_t)+1);              %area of interest in rx samples
%                     [CFO,~]     = est_CFO(aoi,length(sts_t),true_arrival*5-1);                                          %estimate CFO 
                    CFO         = 0;
                    cfos(i,:)   = [CFO,cfos(i,1:end-k)];                                                                %shift current CFO values right by one
                end
            else                                                                                                        %logic for general case (j ~= 1)
                tx_ids(i,:) = [tx_ids(i,1:j-1),(1:shift_right)+(true_arrival-1),tx_ids(i,j:end-shift_right)];           %shift tx_ids right by 'shift_right' number of indicies
                inds(i,:)   = [inds(i,1:j-1), interval*(-1:-1*shift_right)+inds(i,j),inds(i,j:end-shift_right)];        %shift inds right by 'shift right' number of indicies
                for k = 1:shift_right                                                                                   %deal with CFO estimation
%                     aoi         = IQ(i, inds(i,j+k-1)+length(zc)+1 : inds(i,j+k-1)+length(zc)+45*length(sts_t)+1);      %are of interest in rx samples
%                     [CFO,~]     = est_CFO(aoi,length(sts_t),true_arrival*5-1);                                          %estimate CFO
                    CFO         = 0;
                    cfos(i,:)   = [cfos(i,1:j+(k-1)-1),CFO,cfos(i,j:end-k)];                                            %shift current CFO values right by one
                end
            end
        end
        if true_arrival == N_TX && est_arrival < true_arrival                                                           %special logic for when true tx channel is channel N_TX, but estimated detection isn't
            shift_right     = est_arrival;                                                                              %number of spaces to chift tx_ids, inds, and cfos right
            tx_ids(i,:)     = [tx_ids(i,1:j-1),(1:shift_right)+(true_arrival-1),tx_ids(i,j:end-shift_right)];           %shift tx_ids right by 'shift_right' number of indicies
            inds(i,:)       = [inds(i,1:j-1), interval*(-1:-1*shift_right)+inds(i,j),inds(i,j:end-shift_right)];        %shift inds right by 'shift_right' number of indicies
            for k = 1:shift_right                                                                                       %deal with each individual CFO estimation
%                 aoi         = IQ(i, inds(i,j+k-1)+length(zc)+1 : inds(i,j+k-1)+length(zc)+45*length(sts_t)+1);          %are of interest in rx samples
%                 [CFO,~]     = est_CFO(aoi,length(sts_t),true_arrival*5-1);                                              %estimation CFO
                CFO         = 0;
                cfos(i,:)   = [cfos(i,1:j+(k-1)-1),CFO,cfos(i,j:end-k)];                                                %shift current CFO values right by one, insert newly calculated value
            end
        end
    end
end
tx_ids              = tx_ids(:,1:last_idx);
inds                = inds(:,1:last_idx);
cfos                = cfos(:,1:last_idx);

%% Determine number of soundings per channel
num_soundings       = zeros(N_RX,N_TX);                     %matrix to hold number of detections from specific Tx per rx channel
for i = 1:N_RX                                              %for each Rx channel
    for j = 1:N_TX                                          %for each transmit channel
        num_soundings(i,j)   = sum(tx_ids(i,:) == j);       %how many detections were from transmitter j
    end
end
n_soundings         = min(min(num_soundings));              %number of soundings heard





%% ############################################################################
%% PROCESS DATA, FIND CHANNELS, CALCULATE DOA
%% ############################################################################

%% For each transmitter, look at each receiver and calculate channel
H_ests              = zeros(n_soundings,N,N_TX,N_RX);       %allocate space for freq channel ests
h_ests              = zeros(n_soundings,N,N_TX,N_RX);       %allocate space for time channel ests
ang_ESPRIT          = zeros(K,n_soundings-L);               %allocate space for DOA estimates using ESPRIT
ang_APG             = zeros(K,n_soundings-L);               %allocate space for DOA estimates using APG
measurements        = zeros(N_TX,N_RX,n_soundings);         %allocate space for measurements

for j = 1:N_TX
    Rx_t    = zeros(N_RX,N,n_soundings);                    %allocate space for time domain signals between each Rx and Tx j
    for k = 1:N_RX                                          %for each rx channel
        [~, ind]        = find(tx_ids(k,:) == j);           %find start indices of soundings from Tx j on Rx k
        for l = 1:length(ind)                               %for each sounding-
            data_t          = IQ(k,inds(k,ind(l))+length(zc)+45*length(sts_t)-1:inds(k,ind(l))+length(zc)+45*length(sts_t)+N-2);      %get rx data
            
            %% Correct CFO
            if CFO_CORRECTION
                data_t          = data_t.*exp(1i*2*pi*cfos(k,ind(l)).*[0:length(data_t)-1]);%correct CFO
            end
            
            %% Process each data capture
            data_f          = fft(data_t,N);                %fft of capture
            H_est           = data_f./transpose(Pilots);    %channel estimate (freq)
            H_est(GBs)      = 0;                            %guard bands set to zero (otherwise divide by zero error)
            H_ests(l,:,j,k) = H_est;                        %store channel estimate (freq)
            h_est           = ifft(H_est,N);                %channel tap delays
            h_ests(l,:,j,k) = h_est;                        %store channel tap delays
            Rx_t(k,:,l)     = H_est;                       %save raw rx data for DOA processing
            
            %% Check/Correct if start index was off by 1 or more samples
            if SAMPLE_CORRECTION                                        
                [~,arrival_idx] = max(abs(h_est));                      %peak detection of time domain channel response
                if arrival_idx ~= 1                                     %if arrival wasn't on index 1 it must be corrected
                    [~,ind_error]   = max(circshift(abs(h_est),N/2));   %find how far off the peak arrival was. Add circular shift to avoid wraparound errors
                    ind_error       = ind_error-(N/2)-1;                %correct for circular shift
                    data_t          = IQ(k,inds(k,ind(l))+length(zc)+45*length(sts_t)-1+ind_error:inds(k,ind(l))+length(zc)+45*length(sts_t)+N-2+ind_error);      %get data
                    if CFO_CORRECTION
                        data_t          = data_t.*exp(1i*2*pi*cfos(k,ind(l)).*[0:length(data_t)-1]);%correct CFO
                    end
                    data_f          = fft(data_t,N);                    %fft of capture
                    H_est           = data_f./transpose(Pilots);        %channel estimate (freq)
                    H_est(GBs)      = 0;                                %guard bands set to zero (otherwise divide by zero error)
                    H_ests(l,:,j,k) = H_est;                            %store channel estimate (freq)
                    h_est           = ifft(H_est,N);                    %channel tap delays
                    h_ests(l,:,j,k) = h_est;                            %store channel tap delays
                end
                Rx_t(k,:,l)     = H_est;                               %save raw rx data for DOA processing
            end
            
            %% PLOT
            if CHANNEL_PLOT
                figure(3)
                subplot(2,1,1)
                plot(abs(h_est))
                axis([-10,N+10,0,1.25*max(abs(h_est))])
                grid on
                str         = strcat('Channel Est, time, Rx ',num2str(k),', Tx ',num2str(j));
                title(str)
                xlabel('Time sample')
                ylabel('Channel response magnitude')
                pause(.01)
                subplot(2,1,2)
                plot(abs(H_est))
                grid on
                str2        = strcat('Channel Est, freq, Rx ',num2str(k),', Tx ',num2str(j));
                title(str2)
                xlabel('Subcarrier')
                ylabel('Channel magnitude')
            end
        end
    end
    
    for l = 1:n_soundings
        
        %% DOA estimation
        data                    = squeeze(Rx_t(:,:,l));
        measurements(j,:,l)     = transpose(data(:,DOA_subcarrier));
        if l > L
            data                = squeeze(measurements(j,:,l-(L-1):l));
            [ang_est,~,~]       = ESPRIT_ISA( data, 1,1:2:N_RX, 2:2:N_RX);  %estimate DOA via ESPRIT
            ang_ESPRIT(j,l-L)   = ang_est;                                  %store DOA            
        end
    end
end

if PHASE_CORRECTION
    phase_offsets       = zeros(N_TX,N_RX,N);                   %allocate space for rx channel phase correction
    num_it              = 100;                                   %number of iterations
    delta               = 1;                                    %number of snapshots to move per estimate
    for j = 1:N_TX
        Y               = squeeze(measurements(j,:,:));
        Ryy             = Y*Y';
        [U,S,V]         = svd(Ryy);
        Es              = U(:,1);
        A_phase         = rand(N_RX,1);
        for m = 1:num_it
            A_current   = exp(1i*(A_phase));
            A_learned   = Es*pinv(Es)*A_current;
            argA        = unwrap(angle(A_learned));
            [U,S2,V]    = svd(argA);
            A_phase     = U(:,1)*S2(1:1,1:1)*V(:,1:1)';
        end
        A_cal           = exp(1i*(A_phase));
        A_cal           = A_cal.*repmat(conj(A_cal(1,:)),N_RX,1);
        phase_offsets(j,:,i:i+delta-1) = repmat(A_cal(:,1),1,delta);
        if PHASE_PLOT
            a_vec       = [];
            for k = 1:N_RX
                a_vec   = [a_vec;ones(k,1)*A_cal(k,1)];
            end
            figure(6)
            clf
            zplane(a_vec)
            axis(1.15.*[-1,1,-1,1])
            str         = strcat('capture ',num2str(i));
            title(str)
            pause(.001)
        end
        As = reshape(A_cal(:,1),2,N_RX/2);
        Y= Y.*repmat(conj(A_cal(:,1)),1,size(Y,2));
        for l = 1:n_soundings
            
            %% DOA estimation
            if l > L
                data                = Y(:,l-(L-1):l);
                [ang_est,~,~]       = ESPRIT_ISA( data, 1,1:2:N_RX, 2:2:N_RX);  %estimate DOA via ESPRIT
                ang_ESPRIT(j,l-L)   = ang_est;                                  %store DOA     
            end
        end
    end
end



%% Plot DOAs
WWII_badguys   = [1,n_soundings-L,-25,25];          %axis values
figure(5)
clf
title('DOA estimates, ESPRIT, With Calibration')
hold on
plot(ang_ESPRIT')
grid on
xlabel('Capture number')
ylabel('DOA, \circ')
axis(WWII_badguys)




